// Copyright (c) 2011 ActiveState Software Inc.
// See the LICENSE file for licensing information.

var v8dbg = require('./v8dbg'),
    cmdln = require('./cmdln'),
    util = require('util'),
    events = require('events'),
    net = require('net'),
    path = require('path'),
    assert = require('assert'),
    Log = require('log');

var log = new Log('error')

var DBGP_ERROR_CODES = {
  'INVALID_OPTIONS': 3,
  'BREAKPOINT_COULD_NOT_BE_SET': 200,
  'BREAKPOINT_TYPE_NOT_SUPPORTED': 201,
  'BREAKPOINT_INVALID': 202,
  'NO_CODE_ON_BREAKPOINT_LINE': 203,
  'INVALID_BREAKPOINT_STATE': 204,
  'NO_SUCH_BREAKPOINT': 205,
  'ERROR_EVALUATING_CODE': 206,
  'INVALID_EXPRESSION': 207,
  'CANNOT_GET_PROPERTY': 300,
  'STACK_DEPTH_INVALID': 301,
  'CONTEXT_INVALID': 302
};

var V8_SCOPE_TYPES = {
  0: 'Global',
  1: 'Local',
  2: 'With',
  3: 'Closure',
  4: 'Catch'
};

var DBGP_CONTEXT_IDS = {
  // Map DBGP context IDs to V8 scope types, so we can use 0 for locals.
  0: 1, // Local
  1: 2, // With
  2: 3, // Closure
  3: 4, // Catch
  4: 0  // Global
};

var DBGP_TYPEMAP = {
  'undefined': 'undefined',
  'null': 'object',
  'boolean': 'bool',
  'number': 'float',
  'string': 'string',
  'object': 'object',
  'function': 'function'
}

var INTERNAL_GLOBAL_NAMES = [
  "Array", "Boolean", "Buffer", "clearInterval", "clearTimeout", "console",
  "Date", "decodeURI", "decodeURIComponent", "encodeURI",
  "encodeURIComponent", "Error", "escape", "eval", "EvalError", "execScript",
  "Function", "GLOBAL", "global", "Infinity", "isFinite", "isNaN", "JSON",
  "Math", "NaN", "Number", "Object", "parseFloat", "parseInt", "process",
  "RangeError", "ReferenceError", "RegExp", "root", "setInterval",
  "setTimeout", "String", "SyntaxError", "TypeError", "undefined", "unescape",
  "URIError", "v8debug"
];

function Client(engine) {
  net.Socket.call(this);
  this.engine = engine;
  this.setEncoding('ascii');
}

util.inherits(Client, net.Socket);

(function() {
  this._messageChunk = '';

  this.on('data', function(data) {
    log.info('Got a chunk: ' + data);
    this._messageChunk += data;
    var eom, messages = [];
    while ((eom = this._messageChunk.indexOf('\x00')) != -1) {
        messages.push(this._messageChunk.slice(0, eom));
        this._messageChunk = this._messageChunk.slice(eom+1);
    }
    messages.forEach(function(msg) {
        this.engine.handleMessage(msg);
    }.bind(this));
  });

  this.on('closed', function() {
    this.engine.emit('dbgpClosed');
  });

  this.sendPacket = function(xml) {
    var packet = '';
    packet += xml.length + '\x00' + xml + '\x00';
    this.write(packet);
    log.info("Message packet sent: " + packet);
  };
}).apply(Client.prototype);

exports.Client = Client;

function Engine(debuggee, fileUri, port, host) {
  events.EventEmitter.apply(this);

  this.init(debuggee, fileUri, port, host);

  // Wait a bit before connecting, so that v8 has a chance to open a socket.
  // This will emit the 'ready' event when successful.
  setTimeout(function() {
    this.v8Client.connect(v8dbg.port);
  }.bind(this), 100);
}

util.inherits(Engine, events.EventEmitter);

(function() {
  this.init  = function(child, fileUri, port, host) {
    this.fileUri = fileUri;
    this.port = port;
    this.host = host; // may be undefined; that's okay.
    this.status = 'starting';
    this.reason = 'ok';
    this.stdout = 0; // 0-disable, 1-copy data, 2-redirect

    this._initV8Client();
    this._initDBGPClient();
    this._initChild(child);
    assert.ok(this.child);
  };

  this._initV8Client = function() {
    this.v8Client = new v8dbg.Client();

    this.v8Client.once('ready', function(resp) {
      this.v8Client.setExceptionBreak('uncaught', true, function(resp) {
        if (!resp || !resp.success) {
          log.error("Failed to initialize v8 debugger, aborting.");
          this.terminate();
        } else {
          log.info('dbgp ready')
          this.emit('ready');
        }
      }.bind(this))
    }.bind(this));

    this.v8Client.on('break', function(resp) {
      resp = resp.body;
      log.info('V8 Debugger Client ran into breakpoint:\n\t' +
        resp.sourceLine + ':' + resp.sourceColumn + ' ' + resp.sourceLineText);
      for (var p in resp) {
        log.info(p + ": " + resp[p]);
      }
      log.info("resp.script: " );
      for (var p in resp.script) {
        log.info("  " + p + ": " + resp.script[p]);
      }
      this.status = 'break';
      this.reason = 'ok';
      this.emit('v8ExecutionStopped');
    }.bind(this));

    this.v8Client.on('exception', function(resp) {
      // Note that this only works for uncaught exceptions.
      // TODO This doesn't fire on my machine. Looks like a v8 bug to me.
      resp = resp.body;
      log.error('Debuggee process ran into uncaught exception:\n\t' +
        resp.script.name + ':' + resp.sourceLine + ':' + resp.sourceColumn +
        ' ' + resp.sourceLineText);
      this.status = 'break';
      this.reason = 'exception';
      this.emit('v8ExecutionStopped');
    }.bind(this));

    this.v8Client.on('close', function() {
      log.info('V8 Debugger Client closed');
      this.status = 'stopped';
      this.reason = 'ok';
      this.emit('v8ExecutionStopped');
    }.bind(this));

  };

  this._initDBGPClient = function() {
    this.dbgpClient = new Client(this);

    this.dbgpClient.on('connect', function() {
      log.info('DBGP client connected');

      this.dbgpClient.sendPacket(this.makeInit());
    }.bind(this));

    this.dbgpClient.on('close', function() {
      log.info('DBGP client disconnected');
      this.emit('exit');
    }.bind(this));
  };

  this._initChild = function(childProcess) {
    this.child = childProcess; // the debuggee child process

    this.child.on('exit', function() {
      this.status = 'stopping';
      log.info('Child process exited');
      this.emit('v8ExecutionStopped');
    }.bind(this));

    this.child.stdout.on('data', function(data) {
      if (this.stdout== 0)
        process.stdout.write(data);
      else if (this.stdout == 1)
        this.stdoutCopy(data);
      else if (this.stdout == 2)
        this.stdoutRedirect(data);
    }.bind(this));

    this.child.stderr.on('data', function(data) {
      if (this.stderr == 0)
        process.stderr.write(data);
      else if (this.stderr == 1)
        this.stderrCopy(data);
      else if (this.stderr == 2)
        this.stderrRedirect(data);
    }.bind(this));
  };


  /**
   * Features that are supported in feature_get
   */
  this._readOnlyFeatures = {
    'breakpoint_types': 'line', // TODO implement more breakpoint_types
    'data_encoding': 'base64',
    'encoding': 'utf-8',
    'language_supports_threads': '0',
    'language_name': 'Node.js',
    'language_version': '' + process.version,
    'multiple_sessions': '0',
    'protocol_version': '1',
    'supports_async': '1',
    'supports_postmortem': '1',
    // TODO make writable.
    // must either be limited or we should stop printing function definitions
    'max_data': 30
  };

  /**
   * Features that are supported in feature_get and feature_set
   */
  this._writableFeatures = {
    'max_children': 25,
    'max_depth': 0,
    'show_hidden': 0
  };

  this.getFeature = function(name) {
    if (this._readOnlyFeatures.hasOwnProperty(name)) {
      return this._readOnlyFeatures[name];
    } else if (this._writableFeatures.hasOwnProperty(name)) {
      return this._writableFeatures[name];
    } else {
      return undefined;
    }
  };

  this.setFeature = function(name, value) {
    if (this._writableFeatures.hasOwnProperty(name)) {
      this._writableFeatures[name] = value;
      return true;
    } else {
      return false;
    }
  };


  /**
   * ----- DBGP procotol message handlers
   */

  this.on('feature_get', function(args) {
    var response = {};
    var featureName = '' + args['n'];
    var featureValue, message = null;
    if (!featureName) {
      message = this.makeError(DBGP_ERROR_CODES['INVALID_OPTIONS'],
                               'missing parameter');
    } else {
      featureValue = this.getFeature(featureName);
      if (typeof featureValue !== 'undefined') {
        message = featureValue;
        response['supported'] = '1';
      } else {
        response['supported'] = '0';
      }
    }
    this.sendResponse('feature_get', args['i'], response, '' + message);
  });

  this.on('feature_set', function(args) {
    var featureName = '' + args['n'];
    var newValue = args['v'];
    var message = null;
    var response = { featureName: featureName, success: 0 };
    var errorCode, status = 0, success = 0;
    if (!featureName) {
      errorCode = 'INVALID_OPTIONS';
      message = 'missing parameter';
    } else if (this._readOnlyFeatures.hasOwnProperty(featureName)) {
      message = 'Command ' + featureName + ' not modifiable';
      // This is not an error condition -- just let the client know the feature wasn't changed.
    } else if (isNaN(newValue = parseInt(newValue, 10))) {
      // only numerical features at this point
      errorCode = 'INVALID_OPTIONS';
      message = 'invalid value';
    } else if (!this.setFeature(featureName, newValue)) {
      errorCode = 'INVALID_OPTIONS';
      message = 'Unknown feature name';
    } else {
      response['success'] = '1';
    }
    if (errorCode) {
      log.info("feature_set: got message:"
                  + message
                  + ", errorCode:"
                  + errorCode);
      message = this.makeError(DBGP_ERROR_CODES[errorCode],
                               message);
      this.sendUnsuccessfulResponse('feature_set', args['i'],
                                    {'feature_name': args['n']}, message, true);
    } else {
      log.info("feature_set: "
                  + featureName
                  + " succeeded");
      this.sendResponse('feature_set', args['i'], response);
    }
  });

  this.on('status', function(args) {
    this.sendStatusResponse('status', args['i']);
  });

  this.on('run', function(args) {
    log.info('Starting debuggee...');
    this.v8Client.reqContinue(function(resp) {
      if (resp.success && resp.running) {
        log.info('debuggee running!');
        this.status = 'running';
        this.reason = 'ok';
        this.once('v8ExecutionStopped', function() {
          this.sendStatusResponse('run', args['i']);
        }.bind(this));
      } else {
        this.status = 'stopped';
        this.reason = 'error';
        this.sendStatusResponse('run', args['i']);
      }
    }.bind(this));
  });

  this.on('step_into', function(args) {
    this._firstBreak = true;
    this.v8Client.step('in', 1, function(resp) {
      if (resp.success && resp.running) {
        log.info('debuggee running!');
        this.status = 'running';
        this.reason = 'ok';
        this.once('v8ExecutionStopped', function() {
          this.sendStatusResponse('step_into', args['i']);
        }.bind(this));
      } else {
        this.status = 'stopped';
        this.reason = 'error';
        this.sendStatusResponse('step_into', args['i']);
      }
    }.bind(this));
  });

  this.on('step_over', function(args) {
    this.v8Client.step('next', 1, function(resp) {
      if (resp.success && resp.running) {
        this.status = 'running';
        this.reason = 'ok';
        this.once('v8ExecutionStopped', function() {
          this.sendStatusResponse('step_over', args['i']);
        }.bind(this));
      } else {
        this.status = 'stopped';
        this.reason = 'error';
        this.sendStatusResponse('step_over', args['i']);
      }
    }.bind(this));
  });

  this.on('step_out', function(args) {
    this.v8Client.step('out', 1, function(resp) {
      if (resp.success && resp.running) {
        this.status = 'running';
        this.reason = 'ok';
        this.once('v8ExecutionStopped', function() {
          this.sendStatusResponse('step_out', args['i']);
        }.bind(this));
      } else {
        this.status = 'stopped';
        this.reason = 'error';
        this.sendStatusResponse('step_out', args['i']);
      }
    }.bind(this));
  });

  this.on('stop', function(args) {
    log.info('Stopping debuggee...');
    if (this.child) {
      // Kill the debuggee process, tell the IDE that the session stopped.
      this.once('v8ExecutionStopped', function() {
	this.status = 'stopped';
	this.reason = 'ok';
        this.sendStatusResponse('stop', args['i']);
        this.dbgpClient.end();
      }.bind(this));
      this.child.kill();
      this.child = null;
    } else {
      this.sendStatusResponse('stop', args['i']);
    }
  });

  this.on('stack_get', function(args) {
    var stackDepth = args['d'];
    this.v8Client.fullTrace(function(resp) {
      var frames = [];
      if (!resp || !resp.frames) {
        // Internal error
        this.sendUnsuccessfulResponse('stack_get', args['i']);
        return;
      }

      // Collect requested frames
      if (stackDepth >= 0) {
        // Requested a single frame
        for (var i = 0; i < resp.frames.length; i++) {
          var frame = resp.frames[i];
          if (!frame.script || frame.script.isNative) {
            // Note that we currently don't expose node-internal stuff.
            // This might change.
            continue;
          } else if (frame.index == stackDepth) {
            frames.push(frame);
            break;
          }
        }
        if (frames.length != 1) {
          // Requested stack depth is invalid.
          var error = this.makeError(DBGP_ERROR_CODES['STACK_DEPTH_INVALID']);
          this.sendResponse('stack_get', args['i'], null, error, true);
          return;
        }
      } else {
        // Requested a full stack trace
        for (var i = 0; i < resp.frames.length; i++) {
          var frame = resp.frames[i];
          if (!frame.script || frame.script.isNative)
            continue;
          else
            frames.push(frame);
        }
      }

      // Make a DBGP stack trace and send as reponse
      var framesStr = '';
      for (var i = 0; i < frames.length; i++) {
        var frame = frames[i];
        frame.line += 1; // V8 starts indexing at 0
        framesStr += this.makeFrame(frame.index, 'file', frame.script.name,
          frame.line, frame.func.inferredName);
      }
      this.sendResponse('stack_get', args['i'], { 'success': '1' },
                        framesStr, true);
    }.bind(this));
  });


  this.on('breakpoint_set', function(args) {
    // Only partially implemented.
    var type = args['t'],
        state = args['s'],
        filename = args['f'],
        lineno = parseInt(args['n'], 10);

    var target = this._stripUrlPrefix(filename);
    // V8 starts indexing lines at 0:
    lineno -= 1;
    var func = args['m'];
    switch (type) {
      case 'line':
        this.v8Client.setbreakpoint({
            'type': 'script',
            'target': target,
            'line': lineno,
            'column': 0,
            'enabled': (state != 'disabled')
          }, function(res) {
            if (res && res.success && res.body) {
              var id = res.body.breakpoint;
              this.sendResponse('breakpoint_set', args['i'],
                                { 'id': id, 'state': state });
            } else {
              var error = this.makeError(DBGP_ERROR_CODES['BREAKPOINT_COULD_NOT_BE_SET']);
              this.sendResponse('breakpoint_set', args['i'], { 'status': 'disabled' },
                                error, true);
            }
          }.bind(this));
        break;
      default:
        var error = this.makeError(DBGP_ERROR_CODES['BREAKPOINT_TYPE_NOT_SUPPORTED']);
        this.sendResponse('breakpoint_set', args['i'], { 'status': 'disabled' },
                          error, true);
    }
  });

  /*
  this.on('breakpoint_get', function(args) {

  });
  */

  this.on('breakpoint_update', function(args) {
    var breakpointId = parseInt(args['d'], 10),
        state = args['s'],
        lineNo = args['n'],
        hitValue = parseInt(args['h'], 10),
        hitCondition = args['o'];

    if (lineNo || (hitCondition && hitCondition != '>=')) {
      // Unsupported arguments
      sendError.call(this, DBGP_ERROR_CODES['BREAKPOINT_TYPE_NOT_SUPPORTED']);
      return;
    }

    var bpArgs = {};
    bpArgs.breakpoint = breakpointId;
    bpArgs.enabled = state != 'disabled';
    bpArgs.ignoreCount = (hitValue >= 0) ? hitValue : 0;

    this.v8Client.changebreakpoint(bpArgs, function cb(resp) {
      if (resp && resp.success) {
        this.sendResponse('breakpoint_update', args['i'], null, null, true);
      } else {
        sendError.call(this, DBGP_ERROR_CODES['BREAKPOINT_COULD_NOT_BE_SET']);
      }
    }.bind(this));

    function sendError(errorCode) {
      var error = this.makeError(errorCode);
      this.sendResponse('breakpoint_update', args['i'], null, errorCode, true);
    }
  });

  this.on('breakpoint_remove', function(args) {
    var id = args['d'];
    this.v8Client.clearbreakpoint({
      'breakpoint': id
    }, function(res) {
      if (res && res.success && res.body) {
        this.sendResponse('breakpoint_remove', args['i']);
      } else {
        var error = this.makeError(DBGP_ERROR_CODES['BREAKPOINT_COULD_NOT_BE_SET']);
        this.sendResponse('breakpoint_set', args['i'], { 'status': 'disabled' },
                          error, true);
      }
    }.bind(this));
  });

  /*
  this.on('breakpoint_list', function(args) {

  });
  */

  this.on('stdout', function(args) {
    var target = parseInt(args['c'], 10);
    if ([0,1,2].indexOf(target) != -1) {
      this.stdout = target;
      this.sendResponse('stdout', args['i'], { 'success': 1 });
    } else {
      this.sendUnsuccessfulResponse('stdout', args['i']);
    }
  });

  this.on('stderr', function(args) {
    var target = parseInt(args['c'], 10);
    if ([0,1,2].indexOf(target) != -1) {
      this.stderr = target;
      this.sendResponse('stderr', args['i'], { 'success': 1 });
    } else {
      this.sendUnsuccessfulResponse('stderr', args['i']);
    }
  });

  this.on('property_get', function(args) {
    // TODO respect other parameters such as max_children/pages
    // TODO use frameReceiver when asked for `this`
    var propertyName = args['n'];
    var contextId = parseInt(args['c'], 10) || 0;
    var scopeId = this._scopeForContext(contextId);
    var stackDepth = parseInt(args['d'], 10);
    var complexNames = this._parseComplexJSName(propertyName);
    if (!complexNames) {
      sendPropertyError.call(this, 'Invalid property name.');
      return;
    }

    if (!propertyName || !(scopeId >= 0) || !(stackDepth >= 0)) {
      sendError.call(this, DBGP_ERROR_CODES.INVALID_OPTIONS, 'invalid options');
      return;
    }

    // Get the scope that is supposed to contain the property
    this.v8Client.reqFrame(stackDepth, function(resp) {
      if (!resp || !resp.body || !resp.success) {
        sendError.call(this, DBGP_ERROR_CODES.INVALID_OPTIONS, 'invalid stack depth');
        return;
      }

      var frame = resp.body;
      var scopes = frame.scopes || [];
      var scopeNumber;
      scopes.forEach(function(scope) {
        if (scope.type == scopeId)
          scopeNumber = scope.index;
      });

      this.v8Client.reqScope(scopeNumber, stackDepth, function(resp) {
        if (!resp || !resp.body || !resp.success) {
          sendError.call(this, DBGP_ERROR_CODES.INVALID_OPTIONS, 'invalid context');
          return;
        }

        var scope = resp.body.object;
        resp.refs.forEach(function(obj) {
          if (obj.handle == scope.ref)
            scope.value = obj;
        });

        // Search for the property with the base name
        var property;
        var baseName = complexNames.shift();
        if (baseName == 'this') {
          // `this` is frame.receiver, not a scope member (cp. context_get)
          property = { 'name': 'this', 'ref': frame.receiver.ref };
        } else {
          scope.value.properties.forEach(function(propertyRef) {
            if (propertyRef.name == baseName) {
              property = propertyRef
            }
          });
        }

        if (!property || !property.ref) {
          sendError.call(this, DBGP_ERROR_CODES.CANNOT_GET_PROPERTY, 'invalid context');
          return;
        }

        // Recursively get all objects until we have the object that was requested
        lookupObjectRecursively.call(this, complexNames, property, function(finalProperty) {
          var maxDepth = this.getFeature('max_depth');
          this.v8Client.mirrorObjectsRecursively([finalProperty], maxDepth,
            sendPropertiesResponse.bind(this));

          function sendPropertiesResponse() {
            var propertiesStr = '';
            propertiesStr += this.makeProperty(finalProperty, propertyName);
            this.sendResponse('property_get', args['i'], null, propertiesStr, true);
          }
        }.bind(this));
      }.bind(this));
    }.bind(this));

    function lookupObjectRecursively(names, baseRef, cb) {
      this.v8Client.inlineObjectRef(baseRef, function(property) {
        var child, childName;
        if (!(childName = names.shift())) {
          cb.call(this, property);
          return;
        }

        var children = property.value.properties || [];
        children.forEach(function(propertyHandle) {
          if (propertyHandle.name == childName)
            child = propertyHandle;
        });

        if (!child || !child.ref) {
          sendPropertyError.call(this, 'Property does not exist.');
          return;
        }

        lookupObjectRecursively.call(this, names, child, cb);
      }.bind(this));
    }

    function sendError(code, message) {
      var error = this.makeError(code, message);
      this.sendUnsuccessfulResponse('property_get', args['i'], null, error, true);
    }
  });

  this.on('context_names', function(args) {
    // Hard-wired to always send all of V8's scope names
    var contexts = '';
    for (var id in DBGP_CONTEXT_IDS) {
      var v8Id = DBGP_CONTEXT_IDS[id];
      var name = V8_SCOPE_TYPES[v8Id];
      if (typeof name !== 'string') {
        log.error("DBGP context with id " + id + " is not a V8 scope.")
      }
      contexts += this.makeContext(id, name);
    }
    this.sendResponse('context_names', args['i'], null, contexts, true);
  });

  this.on('context_get', function(args) {
    var contextId = parseInt(args['c'], 10) || 0;
    var scopeId = this._scopeForContext(contextId);
    var stackDepth = parseInt(args['d'], 10);
    this.v8Client.reqFrame(stackDepth, function(resp) {
      if (!resp || !resp.success || !resp.body) {
        sendErrorResponse.call(this, DBGP_ERROR_CODES.INVALID_OPTIONS, 'stack depth invalid');
        return;
      }

      var frame = resp.body || {};
      var scopes = frame.scopes || [];
      var scopeNumber;
      scopes.forEach(function(scope) {
        if (scope.type == scopeId) {
          scopeNumber = scope.index;
          log.debug("Found scope, number " + scopeNumber);
        }
      }.bind(this));

      if (typeof scopeNumber === "undefined") {
        // Probably temporarily not applicable, like "with" or "catch" scopes
        this.sendResponse('context_get', args['i'], null, '', false);
        return;
      }

      this.v8Client.reqScope(scopeNumber, stackDepth, function(resp) {
        if (!resp || !resp.success || !resp.body) {
          sendErrorResponse.call(this, DBGP_ERROR_CODES.INVALID_OPTIONS, 'context id invalid');
          return;
        }

        var scope = resp.body.object;
        resp.refs.forEach(function(obj) {
          if (obj.handle == scope.ref)
            scope.value = obj;
        });

        // If the local scope was requested, include a reference to the frame's
        // receiver object as `this`.
        if (scopeId == 1) { // local
          var thisRef = {"name": "this", "ref": frame.receiver.ref};
          scope.value.properties.push(thisRef);
        }

        if (scopeId == 0) { // global
          this._filterGlobalScope(scope);
        }

        var maxDepth = this.getFeature('max_depth');
        this.v8Client.mirrorObjectsRecursively([scope], maxDepth,
          sendContextResponse.bind(this));

        function sendContextResponse() {
          var propertiesStr = '';
          propertiesStr = scope.value.properties.map(function(property) {
            return this.makeProperty(property, null);
          }, this);
          propertiesStr = propertiesStr.join('');
          this.sendResponse('context_get', args['i'], null, propertiesStr, true);
        }
      }.bind(this));
    }.bind(this));

    function sendErrorResponse(code, message) {
      var error = this.makeError(code, message);
      this.sendUnsuccessfulResponse('context_get', args['i'], null, error, true);
    }
  });

  this.on('source', function(args) {
    var filename = args['f'];
    filename = filename.replace(/^file:\/\/|^dbgp:\/\//, '');
    // Split in two requests- first find everything that contains the filename,
    // and pick the one that has the exact equal filename. Then request that
    // script with the complete source.
    this.v8Client.reqScripts(false, false, true, null, false, filename, function(resp) {
      if (!resp || !resp.success || !resp.body) {
        this.sendUnsuccessfulResponse('source', args['i']);
        return;
      }

      var script, scripts = resp.body;
      for (var i = 0; i < scripts.length; i++) {
        if (scripts[i].name == filename) {
          script = scripts[i];
          break;
        }
      }

      if (!script) {
        this.sendUnsuccessfulResponse('source', args['i']);
        return;
      }

      this.v8Client.reqScripts(false, false, true, [script.id], true, null, function(resp) {
        if (!resp || !resp.success || !resp.body) {
          this.sendUnsuccessfulResponse('source', args['i']);
          return;
        }

        script = resp.body[0];
        this.sendResponse('source', args['i'], { success: 1}, script.source, false);
      }.bind(this));
    }.bind(this));
  });

  this.on('typemap_get', function(args) {
    var response = '';
    for (var typeName in DBGP_TYPEMAP) {
      var typeValue = DBGP_TYPEMAP[typeName];
      response += this.makeType(typeName, typeValue);
    }
    this.sendResponse('typemap_get', args['i'], null, response, true);
  });


  /**
   * ----- Helper functions
   */

  /**
   * Make a <response> message that contains the given command and transaction
   * id, as well as the other given parameters, and message (which can be an
   * xml string or data for a CDATA entity). Then send it.
   */
  this.sendResponse = function(cmd, transactionId, parameters, message, xml) {
    var response = this.makeResponse(cmd, transactionId, parameters, message, xml);
    this.dbgpClient.sendPacket(response);
  };

  /**
   * `sendResponse()` with success="0".
   */
  this.sendUnsuccessfulResponse = function(cmd, transactionId, parameters, message, xml) {
    if (!parameters)
      parameters = {};
    parameters['success'] = 0;
    this.sendResponse(cmd, transactionId, parameters, message, xml);
  };

  /**
   * `sendResponse()` with status=`this.status` and reason=`this.reason`.
   */
  this.sendStatusResponse = function(cmd, transactionId, parameters, message, xml) {
    if (!parameters)
      parameters = {};
    parameters.status = this.status;
    parameters.reason = this.reason;
    this.sendResponse(cmd, transactionId, parameters, message);
  };

  /**
   * Make a <response> message with the given parameters and message. The
   * message can be an xml string or data, which will be base64-encoded.
   */
  this.makeResponse = function(cmd, transactionId, parameters, message, xml) {
    if (!parameters)
      parameters = {};
    parameters['command'] = cmd;
    parameters['transaction_id'] = transactionId;
    if (this.getFeature('data_encoding') == 'base64')
      parameters['encoding'] = 'base64';
    message = message || '';
    var response = '<?xml version="1.0" encoding="UTF-8"?>';
    response += '<response xmlns="urn:debugger_protocol_v1" ';
    for (var name in parameters) {
      // FIXME needs escaping
      var value = parameters[name];
      response += name + '="' + value + '" ';
    }
    response = response.trimRight();
    if (message) {
      response += ">";
      if (xml) {
        response += message;
      } else {
        response += this._encodeData(message);
      }
      response += '</response>';
    } else {
      response += " />";
    }
    return response;
  };

  /**
   * Encodes data according to the current 'data_encoding' feature value.
   */
  this._encodeData = function(data) {
    var dataEncoding = this.getFeature('data_encoding');
    if (dataEncoding == 'none')
      return data;
    else if (dataEncoding == 'base64')
        return (new Buffer('' + data)).toString('base64');
    else
      throw 'Feature data_encoding has an unsupported value!';
  };

  /**
   * Generate an <init> message.
   */
  this.makeInit = function() {
    var ideKey = process.env['IDE_KEY'] || '';
    var dbgpCookie = process.env['DBGP_COOKIE'] || '';
    var appId = process.env['APPID'] || '';
    var init = '<?xml version="1.0" encoding="UTF-8"?>';
    init += '<init xmlns="urn:debugger_protocol_v1" appid="dbgp.js" ' +
               'idekey="'+ ideKey + '" ' +
               'session="'+ dbgpCookie + '" ' +
               'thread="1" ' + // ?
               'parent="' + appId + '" ' + // ?
               'language="node.js" ' +
               'protocol_version="1.0" ' +
               'fileuri="' + this.fileUri + '"/>'; // TODO
    return init;
  };

  /**
   * Generate an <error> message with the given error code message text.
   */
  this.makeError = function(code, message) {
    var error = '<error code="' + code + '">' +
            '<message><![CDATA[' + (message || '') + ']]></message>' +
          '</error>';
    return error;
  };

  /**
   * Generate a (misnamed) <stack> element for stacktraces.
   */
  this.makeFrame = function(level, type, filename, lineno, where, cmdbegin,
                            cmdend) {
    var params = { 'level': level, 'type': type, 'filename': filename,
      'lineno': lineno};
    if (where)
      params['where'] = where;
    if (cmdbegin)
      params['cmdbegin'] = cmdbegin;
    if (where)
      params['where'] = where;
    // Why is <stack> not called <frame>?
    var frame = this._makeXmlTag('stack', params, true);
    return frame;
  };

  /**
   * Generate a <breakpoint> element.
   */
  this.makeBreakpoint = function(id, params, expression) {
    params['id'] = id;
    var bp = this._makeXmlTag('breakpoint', params, false);
    if (expression) {
      bp += this._makeXmlTag('expression');
      bp += '<![CDATA[' + expression + ']]>';
      bp += '</expression>';
    }
    bp += '</breakpoint>';
    return bp;
  };

  /**
   * Convert a given V8 object into a DBGP property XML string recursively.
   */
  this.makeProperty = function(property, fullNameOpt) {
    // Ignore property references that haven't been looked up yet.
    // Also ignore properties with names that start with an underscore.
    if (!property || !property.value || this._shouldIgnoreProperty(property)) {
      return '';
    }

    var propArgs = {}, value;
    propArgs.name = property.name || '';
    propArgs.fullname = propArgs.name;
    if (fullNameOpt) {
      propArgs.fullname = fullNameOpt;
    }
    propArgs.type = property.value.type || '';
    var maxData = this.getFeature('max_data') || -1;
    switch (propArgs.type) {
      case 'undefined':
      case 'null':
        value = ''
        break;
      case 'boolean':
      case 'number':
        // TODO Should we slice numbers?
        value = '' + property.value.value;
        break;
      case 'string':
        value = ('' + property.value.value).slice(0, maxData);
        break;
      case 'function':
        value = '#<Function>';
        break;
      case 'object':
        value = ('' + property.value.text).slice(0, maxData);
        break;
      default:
        value ='';
        break;
    }

    propArgs.children = '0';
    var children = [];
    if (propArgs.type == 'object') {
      if (property.value.properties && property.value.properties.length) {
        var numChildren = 0;
        // Recurse over child properties to get XML child elements.
        propArgs.children = '1';
        children = property.value.properties.map(function(child) {
          if (!this._shouldIgnoreProperty(child))
            // Komodo wants the precise value of numchildren
            numChildren ++;
          var fullname = propArgs.fullname + '["' + child.name + '"]';
          return this.makeProperty(child, fullname);
        }.bind(this));
        propArgs.numchildren = numChildren
        children = children.join('');
      }
    }
    var propXml = this._makeXmlTag('property', propArgs, false);
    propXml += '<![CDATA[' + value + ']]>';
    if (children)
      propXml += children;
    propXml += '</property>';
    return propXml;
  };

  this.makeContext = function(id, name) {
    return this._makeXmlTag('context', {
      'id': id,
      'name': name
    }, true);
  };

  this.makeType = function(name, type) {
    return this._makeXmlTag('map', {
      'name': name,
      'type': type
    }, true);
  };

  /**
   * Send data as a dbgp stdout message and to our stdout file.
   */
  this.stdoutCopy = function(data) {
    assert.equal(this.stdout, 1);
    var msg = this._makeStreamMessage('stdout', data);
    this.dbgpClient.sendPacket(msg);
    process.stdout.write(data);
  };

  /**
   * Send data as a dbgp message.
   */
  this.stdoutRedirect = function(data) {
    assert.equal(this.stdout, 2);
    var msg = this._makeStreamMessage('stdout', data);
    this.dbgpClient.sendPacket(msg);
  };

  /**
   * Send data as a dbgp stderr message and to our stderr file.
   */
  this.stderrCopy = function(data) {
    assert.equal(this.stderr, 1);
    var msg = this._makeStreamMessage('stderr', data);
    this.dbgpClient.sendPacket(msg);
    process.stderr.write(data);
  };

  /**
   * Send data as a dbgp stderr message.
   */
  this.stderrRedirect = function(data) {
    assert.equal(this.stderr, 2);
    var msg = this._makeStreamMessage('stderr', data);
    this.dbgpClient.sendPacket(msg);
  };

  this._makeStreamMessage = function(type, data) {
    var msg = '<?xml version="1.0" encoding="UTF-8"?>';
    msg += this._makeXmlTag('stream', { 'type': type }, false);
    msg += this._encodeData(data);
    msg += '</stream>';
    return msg;
  };

  this._makeXmlTag = function(tagName, params, closing) {
    var tag = '<' + tagName + ' ';
    for (var name in params) {
      var value = params[name];
      tag += name + '="' + this._xmlEscape(value) + '" ';
    }
    tag.trimRight();
    tag += closing ? '/>' : '>';
    return tag;
  };

  var _escapes = {
    "<": "&gt;",
    ">": "&lt;",
    "&": "&amp;",
    "'": "&apos;",
    '"': "&quot;"
  };
  var _escapableChars_re = /([<>&\'\"])/g;
  var _xmlEscapeReplFunc = function(c) {
    return _escapes[c];
  };
  this._xmlEscape = function(s) {
    try {
      return s.replace(_escapableChars_re, _xmlEscapeReplFunc)
    } catch(ex) {
      return s;
    }
  };

  /**
   * Takes a JavaScript name like `'this.foo['bar']'` and returns a list of its
   * components (`['this', 'foo', 'bar']`). The name string is not checked for
   * sanity.
   */
  this._parseComplexJSName = function(nameStr) {
    if (typeof nameStr !== 'string')
      return;

    var names = [];
    var firstPartRe = /^([^\.\[]+)/;
    var nthPartRe = /^\[\'([^\']+)\'\]|^\[\"([^\"]+)\"\]|^\.([^\.\[]+)/;
    var regEx, match, name;

    while (true) {
      // Use firstPartRe for the first match, then nthPartRe.
      regEx = regEx ? nthPartRe : firstPartRe;
      if (!(match = regEx.exec(nameStr)))
        break;

      for (var i = 1; i < match.length; i++) {
        if (match[i])
          break;
      }
      if (i == match.length)
        break; // we shouldn't get here
      names.push(match[i]);
      nameStr = nameStr.slice(match[0].length);
    }

    if (nameStr.length)
      return; // nameStr could not be matched entirely.
    else
      return names;
  };

  this._filterGlobalScope = function(scope) {
    // Node and JS internals in the global scope are usually not that
    // interesting, but displaying them with every context_get can slow things
    // down a lot. So we hide them if the Server doesn't tell us otherwise.
    if (this.getFeature('show_hidden'))
      return;

    var property;
    scope.value.properties = scope.value.properties.filter(function(property) {
      return INTERNAL_GLOBAL_NAMES.indexOf(property.name) == -1;
    });
  };

  this._shouldIgnoreProperty = function(property) {
    return !this.getFeature('show_hidden') && property.name[0] == "_";
  };

  this._scopeForContext = function(contextId) {
    // Return a V8 scope ID for a DBGP context ID.
    return DBGP_CONTEXT_IDS[contextId];
  };

  /**
   * Removes a leading 'file://' or 'dbgp:///' from url.
   */
  this._stripUrlPrefix = function(url) {
    var re = /^(file|dbgp)\:\/*/
    url = url.replace(re, '/');
    return url;
  };

  this._parseArgs = function(args) {
    if (args.length % 2 != 0)
      throw "Invalid arguments format!";
    var parsedArgs = {};
    for (var i=0; i < args.length; i += 2) {
      var name = args[i];
      if (name[0] != '-')
        throw "Invalid argument name: " + name;
      name = name.slice(1);
      var value = args[i+1];
      parsedArgs[name] = value;
    }
    return parsedArgs;
  };

  this.handleMessage = function(msg) {
    log.info("Got message: " + msg);
    var args = cmdln.parseArgv(msg); // TODO try-catch
    var cmd = args.shift();
    this.handleCommand(cmd, args);
  };

  this.handleCommand = function(cmd, args) {
    log.info("Got command: " + cmd);
    args = this._parseArgs(args);
    if (this.listeners(cmd).length > 0) {
      // TODO Potentially bad because we also have other event listeners.
      // It's probably a good idea to separate protocol implementation from the
      // rest of the Enine.
      this.emit(cmd, args);
    } else {
      log.warning("Unsupported command " + cmd);
      this.sendUnsuccessfulResponse(cmd, args['i']);
    }
  };

  /**
   * Initiates the debugging session. Call this, once the Engine emits 'ready'.
   * The dbgp client will connect to the dbgp server (Komodo) to initiate the
   * session for the user.
   */
  this.connect = function() {
    this.dbgpClient.connect(this.port, this.host);
  };

  /**
   * Terminates the debugging session by closing the connection to the dbgp
   * server (Komodo).
   */
  this.terminate = function() {
    this.dbgpClient.end();
  };
}).apply(Engine.prototype);

exports.Engine = Engine;

function capitalize(str) {
  str = str.toLowerCase();
  str[0] = str[0].toUpperCase();
  return str;
}
