// Copyright (c) 2011 ActiveState Software Inc.
// See the LICENSE file for licensing information.

var _debugger = require('./_debugger'),
  util = require('util');

/**
 * This extends node's implementation of the v8 debugger protocol (_debugger.js)
 */
function Client() {
  _debugger.Client.apply(this);
}

util.inherits(Client, _debugger.Client);

exports.port = _debugger.port;
exports.Client = Client;

Client.prototype._onResponse = function(res) {
  // FIXME fix for https://github.com/joyent/node/issues/717
  // FIXME fix for missing 'exception' event
  
  for (var i = 0; i < this._reqCallbacks.length; i++) {
    if (this._reqCallbacks[i].request_seq == res.body.request_seq) {
      var cb = this._reqCallbacks[i];
      break;
    }
  }

  var self = this;
  var handled = false;

  if (res.headers.Type == 'connect') {
    // Request a list of scripts for our own storage.
    self.reqScripts();
    self.emit('ready');
    handled = true;

  } else if (res.body && res.body.event == 'break') {
    this.emit('break', res.body);
    handled = true;

  } else if (res.body && res.body.event == 'exception') {
    this.emit('exception', res.body);
    handled = true;

  } else if (res.body && res.body.event == 'afterCompile') {
    this._addHandle(res.body.body.script);
    handled = true;

  } else if (res.body && res.body.event == 'scriptCollected') {
    // ???
    this._removeScript(res.body.body.script);
    handled = true;

  }

  if (cb) {
    this._reqCallbacks.splice(i, 1);
    handled = true;
    cb(res.body);
  }

  if (!handled) this.emit('unhandledResponse', res.body);
};

Client.prototype.setbreakpoint = function(arguments, cb) {
  var req = {
    'command': 'setbreakpoint',
    'arguments': arguments
  };
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.clearbreakpoint = function(arguments, cb) {
  var req = {
    'command': 'clearbreakpoint',
    'arguments': arguments
  };
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.changebreakpoint = function(arguments, cb) {
  var req = {
    'command': 'changebreakpoint',
    'arguments': arguments
  };
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

/**
 * Inlines the value of a given object reference by doing a reqLookup.
 */
Client.prototype.inlineObjectRef = function(object, cb) {
  this.reqLookup([object.ref], function(resp) {
    if (!resp || !resp.success || !resp.body) {
      console.error('problem with reqLookup');
      if (cb) cb(object);
      return;
    }
    
    object.value = resp.body[object.ref];
    if (cb) cb(object);
  });
};

/**
 * Inlines the property references of a given Array of objects. Returns an Array
 * cotaining the inlined property objects.
 */
Client.prototype.mirrorObjects = function(objects, cb) {
  // Get all refs to look up
  var lookupRefs = [];
  objects.forEach(function(object) {
    object = object.value;
    if (object.type != 'object')
      return;

    lookupRefs = lookupRefs.concat(object.properties.map(function(property) {
      return property.ref;
    }));
  }, this);
  
  if (lookupRefs.length == 0) {
    if (cb) cb(objects);
    return;
  }
  
  this.reqLookup(lookupRefs, function(resp) {
    if (!resp || !resp.success || !resp.body) {
      console.error('problem with reqLookup');
      if (cb) cb(objects);
      return;
    }
    
    // Replace references with mirror objects
    var properties = [];
    objects.forEach(function(object) {
      object = object.value;
      if (object.type != 'object')
        return;
      
      for (var i = 0; i < object.properties.length; i++) {
        var propertyRef = object.properties[i];
        var mirror = resp.body[propertyRef.ref];
        if (mirror) {
          propertyRef.value = mirror;
          properties.push(propertyRef);
        } else {
          console.error('Could not find property: ' + propertyRef.name);
        }
      }
    });
    
    if (cb) cb(properties);
  }.bind(this));
};

/**
 * Inlines the properties of a given Array of objects recursively, up to a 
 * given maximum depth.
 */
Client.prototype.mirrorObjectsRecursively = function(objects, maxDepth, cb) {
  var depth = 0;
  recurse.call(this, objects);
  function recurse(objects_) {
    if (depth++ > maxDepth) {
      if (cb) cb(objects_);
      return;
    }
    this.mirrorObjects(objects_, arguments.callee.bind(this))
  }
};

Client.prototype.reqScope = function(scopeNumber, frameNumber, cb) {
  var req = {
    'command': 'scope',
    'arguments': {
      'number': scopeNumber
    }
  };
  if (typeof frameNumber !== 'undefined' && frameNumber != null)
    req.arguments.frameNumber = frameNumber;
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

/*
DON'T use this. The inlined properties are *not* complete.
Client.prototype.reqScopeInline = function(scopeNumber, frameNumber, cb) {
  var req = {
    'command': 'scope',
    'arguments': {
      'number': scopeNumber,
      'inlineRefs': true
    }
  };
  if (typeof frameNumber !== 'undefined' && frameNumber != null)
    req.arguments.frameNumber = frameNumber;
  console.log(scopeNumber + JSON.stringify(req));
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};
*/

Client.prototype.reqScopes = function(frameNumber, cb) {
  var req = {
    'command': 'scopes',
    'arguments': {}
  };
  if (typeof frameNumber !== 'undefined' && frameNumber != null)
    req.arguments.frameNumber = frameNumber;
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.reqScopesInline = function(frameNumber, cb) {
  var req = {
    'command': 'scopes',
    'arguments': {
      'inlineRefs': true
    }
  };
  if (typeof frameNumber !== 'undefined' && frameNumber != null)
    req.arguments.frameNumber = frameNumber;
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.reqScopeHandles = function(frameNumber, scopeType, cb) {
  // Get all scopes
  this.reqScopes(frameNumber, function(resp) {
    if (!resp || !resp.success){
      console.error('Failed to get scopes');
      return [];
    }
    
    // Build a list of scopes by comparing type with the given contextId and
    // fetching the handle of matching scopes from the refs section
    var scopes = [];
    resp.body.scopes.forEach(function(scope) {
      if (typeof scopeType == 'number' && scope.type != scopeType)
        return;
      
      var handleId = scope.object.ref;
      var scopeObject;
      for (var i = 0; i < resp.refs.length; i++) {
        scopeObject = resp.refs[i];
        if (scopeObject.handle == handleId)
          break;
      }
      if (!scopeObject)
        return;
      scopes.push(scopeObject);
    }, this);
    
    if (cb) cb(scopes);
  });
};

Client.prototype.reqFrame = function(number, cb) {
  var req = {
    'command': 'frame',
    'arguments': {}
  };
  if (typeof number !== 'undefined' && number != null)
    req.arguments.number = number;
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.reqFrameInline = function(number, cb) {
  var req = {
    'command': 'frame',
    'arguments': { 'inlineRefs': true }
  };
  if (typeof number !== 'undefined' && number != null)
    req.arguments.number = number;
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.setExceptionBreak = function(type, enabled, cb) {
  var req = {
    'command': 'setexceptionbreak',
    'arguments': {}
  };
  req.arguments.type = type;
  if (typeof enabled == 'boolean')
    req.arguments.enabled = enabled; // toggles otherwise
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};

Client.prototype.reqScripts = function(native_, extension, normal, ids, 
                                       includeSource, filter, cb) {
  var req = {
    'command': 'scripts',
    'arguments': {}
  };
  var types = 0; // Bit mask that determines which types of scripts are returned
  if (native_)   types |= 1;
  if (extension) types |= 2;
  if (normal)    types |= 4;
  req.arguments.types = types || 4; // request normal scripts if nothing else
  
  if (ids)
    req.arguments.ids = ids;
  req.arguments.includeSource = !!includeSource;
  if (typeof filter == 'number' || typeof filter == 'string')
    req.arguments.filter = filter;
  
  this.req(req, function(res) {
    if (cb) cb(res);
  });
};
